from unittest import result
import mugrade
import numpy as np

def objects_equal(value,ref):
    """ Test if two objects are equal according to our autograding rules.
    We have to do this manually so that numpy arrays are compard properly. """

    try:
        if isinstance(ref, dict):
            if sorted(value.keys()) != sorted(ref.keys()):
                return False
            for k,v in ref.items():
                if not objects_equal(v, value[k]):
                    return False
            return True
    
        elif isinstance(ref, list) or isinstance(ref, tuple):
            if len(value) != len(ref):
                return False
            for a,b in zip(ref, value):
                if not objects_equal(a, b):
                    return False
            return True
    
        elif isinstance(ref, np.ndarray):
            if value.shape != ref.shape:
                return False
            if not np.allclose(ref, value, atol=1e-8):
                return False
            return True
          
        elif isinstance(ref, float) or isinstance(ref, np.float32):
          return np.allclose(value, ref)
    
        else:
            return value == ref
    except:
        return False

def test_process_movies():
    X_train, X_test, movies = process_movies(*read_csv()[0], read_csv()[1])
    with mugrade.test:
        assert X_train.nnz == 90752
    with mugrade.test:
        assert X_train.shape == (610, 9724)
    with mugrade.test:
        assert objects_equal(X_train.data.mean(), 3.502870460155148)
    with mugrade.test:
        assert len(movies) == 9724

def submit_process_movies():
    X_train, X_test, movies = process_movies(*read_csv()[0], read_csv()[1])
    mugrade.submit(X_test.nnz)
    mugrade.submit(X_test.shape)
    mugrade.submit(X_test.data.mean())

def test_cf_error():
    with mugrade.test:
        assert objects_equal(cf_error(np.array([[5., 0]]), np.array([[.1, .2, .3, .4, .5]]), np.array([[1., 2., 3., 4., 5.], [3., 4., 5., 6., 7.]]), 0), 0.25)
    X_train, X_test, movies = process_movies(*read_csv()[0], read_csv()[1])
    with mugrade.test:
        assert objects_equal(cf_error(X_train.toarray(), np.ones((610,3)), np.ones((9724,3)), 0.5), 1.0844141175952045)

def submit_cf_error():
    mugrade.submit(cf_error(np.array([[3., 0]]), np.array([[.1, .2, .3, .4, .5]]), np.array([[1., 2., 3., 4., 5.], [3., 4., 5., 6., 7.]]), 1))
    X_train, X_test, movies = process_movies(*read_csv()[0], read_csv()[1])
    mugrade.submit(cf_error(X_test.toarray(), np.ones((610,3)), np.ones((9724,3)), 0.0))

def test_cf_train():
    X = np.array([[0, 5.0, 0], [0, 0, 2.0], [3.0, 5.0, 0]])
    U, V, b = cf_train(X, 2, niters=1, lam=0)
    with mugrade.test:
        assert objects_equal(U, np.array([[0.5747707969110369, 1.5356451563687277], [0.4117714360839387, 0.013553397019259556], [-0.1513604990615858, -0.19178485493262803]]))
    with mugrade.test:
        assert objects_equal(V, np.array([[2.1477947188552027, 2.2155488743962866], [-17.669577994974066, 7.427469411113261], [-2.8037011156399956, -43.93850223796603]]))
    with mugrade.test:
        assert objects_equal(b, 3.75)
    with mugrade.test:
        assert objects_equal(U@V.T + b, np.array([[8.386786579822566, 5.0, -65.3354336643658], [4.664428729306772, -3.4251600642104085, 2.0], [2.999999999999999, 5.0, 12.60110887774806]]))

    X = np.array([[0, 5.0, 0], [0, 0, 2.0], [3.0, 5.0, 0]])
    U, V, b = cf_train(X, 2, niters=10, lam=1e-3)
    with mugrade.test:
        assert objects_equal(U, np.array([[-0.15354880768286722, 0.008148303407109314], [0.2298278400704812, 0.0356685345046495], [-0.16609929499112405, -0.22831177170256511]]))
    with mugrade.test:
        assert objects_equal(V, np.array([[1.5792759778637246, 2.135573327887181], [-8.117750766043445, 0.43078091730180024], [-7.435176787526984, -1.1539152946546627]]))
    with mugrade.test:
        assert objects_equal(b, 3.75)
    with mugrade.test:
        assert objects_equal(U@V.T + b, np.array([[3.524905356020586, 4.999981084808821, 4.882260078709149], [4.18913435780057, 1.8996801992224648, 2.000030910877661], [3.000106843389752, 5.000000324708643, 5.248430067859896]]))

    X_train, X_test, movies = process_movies(*read_csv()[0], read_csv()[1])
    with mugrade.test:
        assert objects_equal(cf_error(X_train.toarray(), *cf_train(X_train.toarray(), 3, niters=10)), 0.6646328954115686)
    with mugrade.test:
        assert objects_equal(cf_error(X_test.toarray(), *cf_train(X_train.toarray(), 2, niters=10)), 0.817086459822611)
    with mugrade.test:
        assert objects_equal(cf_error(X_test.toarray(), *cf_train(X_train.toarray(), 5, lam=5.0, niters=4)), 0.82242863311804)

def submit_cf_train():
    X_train, X_test, movies = process_movies(*read_csv()[0], read_csv()[1])
    mugrade.submit(cf_error(X_train.toarray(), *cf_train(X_train.toarray(), 4, niters=8)))
    mugrade.submit(cf_error(X_test.toarray(), *cf_train(X_train.toarray(), 3, niters=7)))
    mugrade.submit(cf_error(X_test.toarray(), *cf_train(X_train.toarray(), 10, lam=5.0, niters=4)))

def test_cf_recommend():
    X = np.array([[0, 5.0, 0], [0, 0, 2.0], [3.0, 5.0, 0]])
    U, V, b = cf_train(X, 3, niters=10, lam=1e-3)
    with mugrade.test:
        assert objects_equal(cf_recommend(X, U, V, b), np.array([2, 0, 2]))

    X_train, X_test, movies = process_movies(*read_csv()[0], read_csv()[1])
    with mugrade.test:
        assert objects_equal(cf_recommend(X_train.toarray(), *cf_train(X_train.toarray(), 4, niters=4))[:10], np.array([277, 1938, 7312, 277, 461, 2224, 277, 257, 277, 1172]))

def submit_cf_recommend():
    X_train, X_test, movies = process_movies(*read_csv()[0], read_csv()[1])
    mugrade.submit(cf_recommend(X_train.toarray(), *cf_train(X_train.toarray(), 3, niters=3))[:10])
